/*
 * Copyright (C) 2023-2026 Luis Chamberlain <mcgrof@kernel.org>
 * Copyright (C) 2023-2026 Colin Ian King.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 */
#include <linux/module.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

extern int finit_module(int fd, const char *param_values, int flags);

int main(void)
{
	const char *module = "hello";
	int fd;

	fd = open(module, O_RDONLY | O_CLOEXEC);
	if (fd > 0) {
		int ret;

		ret = finit_module(fd, "", 0);
		(void)ret;
	}

	return 0;
}
